/**
    @file
    @brief BLE module
    @copyright Copyright (c) 2012 Packet Power LLC. All Rights Reserved.
    @author Paul Bieganski

    This is the ble module which handles:
    - ble initialization routines

*/

#include <string.h>
#include "nordic_common.h"
#include "app_util.h"
#include "p5_svc.h"

/**@brief Connect event handler.
 *
 * @param[in]   p_lbs       LEDButton Service structure.
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 */
static void on_connect(P5Svc_t* const p_p5Svc, ble_evt_t * p_ble_evt)
{
    p_p5Svc->connHandle = p_ble_evt->evt.gap_evt.conn_handle;
}

/**@brief Disconnect event handler.
 *
 * @param[in]   p_lbs       LEDButton Service structure.
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 */
static void on_disconnect(P5Svc_t* const p_p5Svc, ble_evt_t * p_ble_evt)
{
    UNUSED_PARAMETER(p_ble_evt);
    p_p5Svc->connHandle = BLE_CONN_HANDLE_INVALID;
}

void P5SvcOnBleEvt(P5Svc_t* const p_p5Svc, 
                   ble_evt_t* const p_ble_evt)
{
    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GAP_EVT_CONNECTED:
            on_connect(p_p5Svc, p_ble_evt);
            break;
            
        case BLE_GAP_EVT_DISCONNECTED:
            on_disconnect(p_p5Svc, p_ble_evt);
            break;
            
        default:
            break; 
    }
}

static uint32_t ledCharAdd(P5Svc_t* const p_p5Svc)
{
    ble_gatts_char_md_t char_md;
    ble_gatts_attr_md_t cccd_md;
    ble_gatts_attr_t    attr_char_value;
    ble_uuid_t          ble_uuid;
    ble_gatts_attr_md_t attr_md;

    memset(&cccd_md, 0, sizeof(cccd_md));

    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cccd_md.read_perm);
    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cccd_md.write_perm);

    cccd_md.vloc = BLE_GATTS_VLOC_STACK;
    
    memset(&char_md, 0, sizeof(char_md));
    
    char_md.char_props.read   = 1;
    char_md.char_props.notify = 1;
    char_md.p_char_user_desc  = NULL;
    char_md.p_char_pf         = NULL;
    char_md.p_user_desc_md    = NULL;
    char_md.p_cccd_md         = &cccd_md;
    char_md.p_sccd_md         = NULL;
    
    ble_uuid.type = p_p5Svc->uuidType;
    ble_uuid.uuid = P5_BLE_UUID_LED_CHAR;

    memset(&attr_md, 0, sizeof(attr_md));

    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.read_perm);
    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.write_perm);
    
    attr_md.vloc       = BLE_GATTS_VLOC_STACK;
    attr_md.rd_auth    = 0;
    attr_md.wr_auth    = 0;
    attr_md.vlen       = 0;
    
    memset(&attr_char_value, 0, sizeof(attr_char_value));

    attr_char_value.p_uuid       = &ble_uuid;
    attr_char_value.p_attr_md    = &attr_md;
    attr_char_value.init_len     = sizeof(uint8_t);
    attr_char_value.init_offs    = 0;
    attr_char_value.max_len      = sizeof(uint8_t);
    attr_char_value.p_value      = NULL;
    
    return sd_ble_gatts_characteristic_add(p_p5Svc->serviceHandle, 
                                           &char_md,
                                           &attr_char_value,
                                           &p_p5Svc->ledCharHandles);


}

uint32_t BleInit(P5Svc_t* p_p5Svc)
{
    uint32_t errCode;
    ble_uuid_t bleUuid;

    // Initialize service structure
    //
    p_p5Svc->connHandle = BLE_CONN_HANDLE_INVALID;
    // ??? led_write_handler???

    // Add base UUID to softdevice's internal list
    //
    ble_uuid128_t baseUuid = P5_BLE_UUID_BASE;
    errCode = sd_ble_uuid_vs_add(&baseUuid, &p_p5Svc->uuidType);
    if (NRF_SUCCESS != errCode)
    {
        return errCode;
    }

    bleUuid.type = p_p5Svc->uuidType;
    bleUuid.uuid = P5_BLE_UUID_SERVICE;

    errCode = sd_ble_gatts_service_add(BLE_GATTS_SRVC_TYPE_PRIMARY,
                                       &bleUuid, 
                                       &p_p5Svc->serviceHandle);
    if (NRF_SUCCESS != errCode)
    {
        return errCode;
    }

    errCode = ledCharAdd(p_p5Svc);
}

uint32_t BleOnLedChange(P5Svc_t* p_p5Svc, uint8_t ledState)
{
    ble_gatts_hvx_params_t params;
    uint16_t len = sizeof(ledState);
    
    memset(&params, 0, sizeof(params));
    params.type = BLE_GATT_HVX_NOTIFICATION;
    params.handle = p_p5Svc->ledCharHandles.value_handle;
    params.p_data = &ledState;
    params.p_len = &len;
    
    return sd_ble_gatts_hvx(p_p5Svc->connHandle, &params);        
}